#include "triangle.h"
#include <math.h>
#include <float.h>
#include <iostream>

using namespace std;

Triangle::Triangle(const Vector& w1, const Vector& w2, const Vector& w3, const Color& c)
	: Shape(c),	v1(w1), v2(w2), v3(w3)
{
	// bounding sphere is centered at the centroid, and has radius big enough to enclose all
	// three vertices
	boundingSphereCenter.x = (w1.x+w2.x+w3.x) / 3;
	boundingSphereCenter.y = (w1.y+w2.y+w3.y) / 3;
	boundingSphereCenter.z = (w1.z+w2.z+w3.z) / 3;

	float dSq1 = (boundingSphereCenter - w1).magSq(),
		  dSq2 = (boundingSphereCenter - w2).magSq(),
		  dSq3 = (boundingSphereCenter - w3).magSq();

	boundingSphereRadius = sqrt(max(max(dSq1, dSq2), dSq3));

	// calculate forward and reverse transforms (using GL for some of the calculation)
	// (Based on code from Prof. Z)

	float tempMatrix[16]; // GL uses a flat 16-element matrix to represent the 4x4 transform,
						  // ..interpreting it in column-major form
	Vector tempV1(v1), tempV2(v2), tempV3(v3); 	     // We'll operate on a copy of the vertices
	
	glPushMatrix();		                             // Save GL's current state

	// translate the triangle so v1 is at the origin
	tempV2 = tempV2 - tempV1;
	tempV3 = tempV3 - tempV1;
	tempV1 = Vector(0.0, 0.0, 0.0);

	// rotate to align last edge with the x axis
	Vector xAxis(1.0, 0.0, 0.0);
	Vector w(tempV3);   // w is a copy of the last edge (V1 is origin, so last vertex = last edge)
	w.normalize();
	Vector v = w.cross(xAxis);                       // vector of rotation
	v.normalize();
	float phi = (float) acos((double) w.dot(xAxis)); // angle of rotation (radians)
	phi *= 360.f / (2.0f * 3.14159f);                // convert to degrees

	if(!v.magSq())                   // What the heck do we do with a zero vector?
		v = Vector(0,1,1);           // This seems to work fine in my test cases, no idea why

	glLoadIdentity();
	glRotatef(phi, v.x, v.y, v.z);               // Have GL compute the rotation transform
	glGetFloatv(GL_MODELVIEW_MATRIX, tempMatrix);    // Grab a copy of the transform matrix
 
	// update vertices
	tempV1 = transformPt(tempMatrix, tempV1);
	tempV2 = transformPt(tempMatrix, tempV2);
	tempV3 = transformPt(tempMatrix, tempV3);

	// rotate to bring second vertex into y=0 plane
	w = tempV2;                                      // Set w to lie along the first edge
	w.x = 0.0f;                                      // project into the x=0 plane

	float sigma = (float) acos((double) (w.z / sqrt(w.magSq()))); 
	                                                 // angle of rotation (radians)

	sigma *= 360.f / (2.0f * 3.14159f);              // convert to degrees
	if(w.y < 0)                                      // rotate about xAxis or -xAxis?
		xAxis = xAxis * -1.0f;

	glLoadIdentity();
	glRotatef(sigma, xAxis.x, xAxis.y, xAxis.z);     // Have GL compute the rotation transform
	glGetFloatv(GL_MODELVIEW_MATRIX, tempMatrix);    // Grab a copy of the transform matrix

	// update vertices
	tempV1 = transformPt(tempMatrix, tempV1);
	tempV2 = transformPt(tempMatrix, tempV2);
	tempV3 = transformPt(tempMatrix, tempV3);

	// Now collapse all that into a single forward (to standard form) transform, which we
	// store in the object for future use
	glLoadIdentity();
	glRotatef(sigma, xAxis.x, xAxis.y, xAxis.z);
	glRotatef(phi, v.x, v.y, v.z);
	glTranslatef(-v1.x, -v1.y, -v1.z);
	glGetFloatv(GL_MODELVIEW_MATRIX, fwdTransform);

	// ..and a single reverse (back to original coordinates) transform
	glLoadIdentity();
	glTranslatef(v1.x, v1.y, v1.z);
	glRotatef(-phi, v.x, v.y, v.z);
	glRotatef(-sigma, xAxis.x, xAxis.y, xAxis.z);
	glGetFloatv(GL_MODELVIEW_MATRIX, revTransform);

	glPopMatrix();                                   // restore the internal GL matrix
	normal = (v2 - v1).cross(v3 - v1).normalize();
//	normal = reverseTransform(Vector(0,1,0));
//	normal.normalize();
}

Triangle::~Triangle(void)
{
}

void Triangle::draw() const
{
	// set color from color
	glColor3f(color.r, color.g, color.b);
	// set vertices from vectors v1, v2, v3
	glBegin(GL_TRIANGLES);
		glNormal3f(normal.x, normal.y, normal.z);
		glVertex3f(v1.x,v1.y,v1.z);
		glNormal3f(normal.x, normal.y, normal.z);
		glVertex3f(v2.x,v2.y,v2.z);
		glNormal3f(normal.x, normal.y, normal.z);
		glVertex3f(v3.x,v3.y,v3.z);
	glEnd(); //end triangle draw
}
