// The world.  Implemented as a singleton class.

#pragma once

#include "Ball.h"
#include "Triangle.h"
#include "includes.h"
#include "camera.h"
#include "collision.h"
#include "Object.h"
#include "Powerup.h"
#include "Breakable.h"

enum Keys { TURNLEFT, FORWARD, TURNRIGHT, STRAFELEFT, BACKWARDS, STRAFERIGHT, JUMP };
enum GameMode { ARENA, RACE, SURVIVAL, DONE };	 // ARENA - you and a few other balls battle
												 // RACE - get to the end before time runs out
												 // SURVIVAL - better balls keep dropping in, survive as long
												 //			   as you can
												 // DONE - game over

class World
{
public:
	~World(void);

	// constants
	static const int MAX_COMPUTERS;
	static const float BALL_RADIUS;
	static const float ELASTICITY;
	static const float GRAVITY;
	static const float FRICTION;
	static const float ACCELERATION; 
	static const float COMP_ACCELERATION;
	static const float MASS;
	static const float COMP_MASS;
	static const float MAX_ACCEL_SPEED;
	static const float B2B_ELASTICITY; 
	static const float JUMP_VEL;
	static const float YVEL_GROUND;
	static const float TURN_RATE;
	static const int BORDER;

	// return a pointer to the singleton instance of the World
	static World* instance();

	// Loads in a level file
	bool loadFile(char* fileName);

	// Redraw everything in the world
	void draw();

	// Advance a timestep. Return true if at least one collision occured in this time,
	// false otherwise (for testing purposes only; normally nobody else really needs to know
	// if collisions happened or not, since World handles them itself)
	bool advanceTimestep();

	// Set ball velocity to a new value
	// (this is only for the testing harness; will probably remove in final version, since
	//  except for testing purposes, the world should be handling all the velocity updating)
	void setBallVelocity(Vector newVelocity) { playerBall.setVelocity(newVelocity); }

	void setMode(GameMode _mode) { mode = _mode; }

	void setKeys(char* _keys);


	// The member callback functions (see private section below) can't be called directly by
	// glut because of C++ internal stuff.
	// glut can call static functions so we make some here that just call the member ones.
	static void mouseMoveCallback(int newx, int newy)
		{ instance()->mouseMove(newx, newy); }
	static void mousePressCallback(int button, int state, int x, int y)
		{ instance()->mousePress(button, state, x, y); }
	static void keyPressCallback(unsigned char key, int x, int y)
		{ instance()->keyPress(key, x, y); }
	static void keyPressReleaseCallback(unsigned char key, int x, int y)
		{ instance()->keyPressRelease(key, x, y); }
	static void physTimerCallback(int dummy) { instance()->physTimer(); };

	// temporary(?) collide call to allow easier reuse with multiple balls
	void collide(Ball& myBall, float timeStep);

private:
	// Private constructor to enforce singleton constraint
	World(void);

	static World* pinstance;

	Ball playerBall;
	Ball** computerBalls;
	int numComputerBalls;
	Camera theCamera;
	GameMode mode;

//	Triangle** triangles;
//	int numTriangles;
	Object **objects;          // All the objects in our world
	int numObjects;

	char keys[20];

	int numSlowSteps;
	int numStrokes;
	float power;
	float zoomFactor;

	DWORD ticks1;
	DWORD ticks2;
	DWORD startTicks;
	DWORD lastSpawnTicks;
	DWORD DTicks;

	Vector goalNear;   // point of goal rectangle nearest the origin
	Vector goalFar;    // point of goal rectangle farthest from the origin
	Vector playerStart;
	Vector AIStart;
	
	// True if the key corresponding to the move direction is currently held down
	bool moveFwd, moveBack, moveLeft, moveRight, moveJump, turnRight, turnLeft;

	int mousex, mousey;

	// Advance a certain time, where 1.0 is one timestep.  This allows us to finish partial
	// timesteps after bounces (by using time < 1.0).
	bool advanceTime(float timeStep);

	// Implementation of Glut callback functions (see public section above for static wrappers)

	// mouse movement callback (only called while a mouse button is held down)
	void mouseMove(int newx, int newy);
	// Mouse button press callback
	void mousePress(int button, int state, int x, int y);
	// Keyboard button callback
	void keyPress(unsigned char key, int x, int y);
	// Keypress release callback
	void keyPressRelease(unsigned char key, int x, int y);

	// callback for physics simulation
	void physTimer();
};
