#include "collision.h"


bool checkSphereIntersection(Vector centerA, float radiusA,
							 Vector centerB, float radiusB)
{
	// true if distance between centers is <= sum of radii
	// (for efficiency, actually check if distance^2 <= (sum of radii)^2)
	float radSum = radiusA + radiusB;
	return (centerA - centerB).magSq() <= radSum*radSum;
}

Collision* checkFaceCollision(const Vector& pathStart, const Vector& pathEnd,
							  float ballRadius, const Triangle* triangle)
{
	// Transform everything to standard coordinates
	Vector pathStartStd = triangle->forwardTransform(pathStart);
	Vector pathEndStd = triangle->forwardTransform(pathEnd);
	Vector v1Std = triangle->forwardTransform(triangle->getV1());
	Vector v2Std = triangle->forwardTransform(triangle->getV2());
	Vector v3Std = triangle->forwardTransform(triangle->getV3());

	Vector relGrad = pathEndStd - pathStartStd; // relative gradient (3d "slope") of path

	// there's only a possibility of a face collision if we're above the face and moving
	// down towards it...
	if( relGrad.y < -0.0001 && pathStartStd.y >= 0.0)
	{
		// find (x,z) coordinates on the ball's path where y = ballRadius
		// if (a,b,c) is the relative gradient and (x0,y0,z0) is a point on the line,
		// in parametric form, x = x0 + at, y = y0 + bt, z = z0 + ct, so solve for
		// t in ballRadius = y0 + bt, then use this to get x and z

		Vector point;
		Vector pointXZ;

		// If the ball is inside the triangle, move the collision point above the triangle to avoid falling through, 
		// otherwise find the real collision point.
		if(pathStartStd.y < ballRadius && pathStartStd.y > 0)
		{
//			cout << "aha!" << endl;
			point = pathStartStd;
			point.y = ballRadius;
			pointXZ = point;
			pointXZ.y = 0.0;
		} else {
			 
			float t = (ballRadius - pathStartStd.y) * (1.0 / relGrad.y);
			// calculate the collision point and the projection of the collision point to the x-z plane
			point = Vector(pathStartStd.x + relGrad.x * t, ballRadius, pathStartStd.z + relGrad.z * t);
			pointXZ = point;
			pointXZ.y = 0.0;
			if(!((point - pathStartStd).magSq() <= relGrad.magSq()))
			{ 
//				cout << "reached" << endl;
				return NULL;
			}
		}

		// Check that the original point (not projected down to x-z) is on the path first,
		// and if so, check that projected-down point is within the triangle (i.e., for each
		// side, y-component of cross-product of v(i+1) - v(i) and point - v(i) must be positive).
		// Return a collision if all these conditions hold.
		if( (v2Std - v1Std).cross(pointXZ - v1Std).y > 0.0
		   && (v3Std - v2Std).cross(pointXZ - v2Std).y > 0.0
		   && (v1Std - v3Std).cross(pointXZ - v3Std).y > 0.0)
		{
			return new Collision(triangle->reverseTransform(point), triangle);
		}
	}

	return NULL;
}

/************************************************************/
/*															*/
/*	this routine finds the closest point q on the line		*/
/*	segment defined by the points "start" and "end"			*/
/*	to the point "v"										*/
/*															*/
/*	it returns alpha where q=start + alpha * (end-start)	*/
/*															*/
/************************************************************/
				
float linePointMinimumDistance(const Vector& start, const Vector& end, const Vector& v)
{
	// construct a vector along the line
	Vector v1=end-start;
	float lenSq = v1.magSq();
	// check if the line is actually a point
	if (lenSq == 0)
		// if so the point start=end is desired point and alpha=0
		return 0; 

	// construct a vector from the line to the point
	Vector v2=v-start;
	// q=start + alpha * v1 is the closest point on the line to the point v
	// where alph is defined below
	float alpha = v1.dot(v2)/(lenSq);

	// if q is not on the line segment then an endpoint of the line
	// segment is the desired point
	if (alpha<0)
		alpha=0;
	if (alpha>1)
		alpha=1;
	
	return alpha;
}

//******************************************************************//
//																	//
//	This routine finds to closest points p1 & p2 on two	line		//
//	segments, l1 (start1 to end1) and l2 (start2 to end2)			//
//  It computes alpha and beta in [0,1] such that					//
//		p1 = start1 + alpha * (end1-start1)							//
//		p2 = start2 + beta * (end2-start2)							//
//	If the lines are parallel -- which should never result in a		//
//	collision, it returns alpha=beta=-1								//
//																	//
//******************************************************************//	

void lineLineMinimumDistance(const Vector& start1, const Vector& end1,
							 const Vector& start2, const Vector& end2,
							 float& alpha, float& beta)
{
	// find vector along first line
	Vector v1=end1-start1;
	float v1dotv1=v1.dot(v1);
	// check if the line is actually a point
	if (v1dotv1==0) {	
		// first line is really a point
		alpha = 0;
		beta = linePointMinimumDistance(start2, end2, start1);
		return;
	}

	// find vector along second line
	Vector v2=end2-start2;
	float v2dotv2=v2.dot(v2);
	// check if the line is actually a point
	if (v2dotv2==0) {
		// second line is really a point
		alpha = linePointMinimumDistance(start1, end1, start2);
		beta=0;
		return;
	}

	// check if lines are parallel
	float v1dotv2=v2.dot(v1);  // we define both for convenience
	float v2dotv1=v1dotv2;

	if (v2dotv2*v1dotv1 != v2dotv1*v1dotv2) {  // do not parallel case first
		// Find closest point
		Vector s=start2-start1;
		float sdotv1=s.dot(v1);
		float sdotv2=s.dot(v2);

		beta = (sdotv2- sdotv1*v1dotv2/v1dotv1)/(v1dotv2*v2dotv1/v1dotv1 - v2dotv2);
		alpha = (sdotv1 + beta*v2dotv1)/(v1dotv1);

		// if alpha and beta are both on their line segments
		// we are done
		// if not, then at least one of the points we are
		// looking for is an endpoint.  we could test each
		// endpoint to find its closest point on the other
		// line.  this would take four calls to linePointMinimumDistance.
		// if we examine the possible cases, however, we
		// need only compute one call to linePointMinimumDistance.
		// we take that approach.

		// alpha and beta are each in one of three states:
		// 0<= alpha <=1, alpha<0, or alpha>1  
		// 0<= beta <=1,  beta <0, or beta>1
		// thus there are nine states to consider
		// testval represents the states
		int testval = 3*(alpha<0) + 6*(alpha>1) + 1*(beta<0) + 2*(beta>1);

		switch (testval) {
		case 0: {	// alpha and beta are in [0,1]
				return;
				}

		case 1: {	// alpha in [0,1], beta<0
				alpha = linePointMinimumDistance(start1,end1,start2);
				beta=0;
				return;
				}

		case 2: {	// alpha in [0,1], beta>1
				alpha = linePointMinimumDistance(start1,end1,end2);
				beta=1;
				return;
				}

		case 3: {	// alpha < 0, beta in [0,1]
				alpha=0;
				beta=linePointMinimumDistance(start2,end2,start1);
				return;
				}

		case 6: {	// alpha > 1, beta in [0,1]
				alpha=1;
				beta=linePointMinimumDistance(start2,end2,end1);
				return;
				}

		case 5: {  // alpha <0, beta >1
				alpha=0;
				beta=1;
				return;
				}

		case 7: {  // alpha >1, beta <0
				alpha=1;
				beta=0;
				return;
				}

		case 4: {  //alpha < 0, beta < 0
					if ((v1*alpha).magSq() > (v2*beta).magSq()) {
						alpha=0;
						beta=linePointMinimumDistance(start2,end2,start1);
					}
					else {
						beta=0;
						alpha = linePointMinimumDistance(start1,end1,start2);
					}
					return;
				}

		case 8: {  // alpha > 1, beta > 1
					if ((start1+v1*alpha-end1).magSq() > (start2+v2*beta).magSq()) {
						alpha=1;
						beta=linePointMinimumDistance(start2,end2,end1);
					}
					else {
						beta = 1;
						alpha =	linePointMinimumDistance(start1,end1,end2);
					}
					return;
				}
		}
	}
	else { 

		// the lines are parallel
		// this should never result in a collision
		alpha = -1;
		beta=-1;
		return;
	}
}
